"""
    This file was generated by flowtool-versioning.
    Homepage: https://github.com/isnok/py-flowlib
"""

VERSION_INFO = {'vcs_info': {'commit': '433e4a228dd5003fcfaadf1806f0a2be94ef4f59',
              'dirt': '',
              'latest_tag': 'flowtool-githooks-0.7.39',
              'latest_tag_commit': '433e4a228dd5003fcfaadf1806f0a2be94ef4f59',
              'latest_tag_version': '0.7.39',
              'prefix': 'flowtool-githooks-',
              'prefix_tag_distances': {'flowtool-githooks-0.7.36': 20,
                                       'flowtool-githooks-0.7.37': 18,
                                       'flowtool-githooks-0.7.38': 10,
                                       'flowtool-githooks-0.7.39': 0},
              'tag_version': {'normalized': '0.7.39',
                              'release': (0, 7, 39),
                              'version': '0.7.39'}},
 'version': '0.7.39'}

def get_version():
    global VERSION_INFO
    return VERSION_INFO.get('version', '$Format:%d %H$')
